import os, sys
import re
import tempfile

import json, yaml

from .util import read_file
from .serialization import DirectoryWriter, ZipWriter, MarkdownSerializer, HTMLMarkdownSerializer
from . import md_util
from . import markdown_magic_comments
from . import source_code_magic_comments
from . import code_parser
from . import javascript
from .publication_visitor import PublicationVisitor
from .task_visitor import TaskVisitor
from .task_repository import TaskRepository
from .publication_repository import PublicationRepository
from .yaml_specification import YAMLSpecification, YAMLSpecificationVisitor
from .md_content_processor import MDContentProcessor, OutputOrganizerHierarchy, LinkProcessorHTML, ImageProcessorCopy
from . import logger

from .messages import msg
from . import messages


# Prepares a publication in a series of HTML files, ready for publishing on the web
class PublicationVisitorHTML(PublicationVisitor, TaskVisitor):
    # yaml_specification - yaml specification of the publication
    # task_repo, publication_repo - repositories of tasks and publication texts
    # dst - destination to store the resulting files (either a directory path, or a zip file path)
    # langs - programming languages included in the publication
    # html - if True, then the Markdown files are converted to HTML (using pypandoc)
    # join_langs - if True, then solutions for all included languages are stored in a single file
    # standalone - if True, all HTML files are standalone (they include a header and the footer and are ready for viewin)
    # css - link to the css file used to style the resulting HTML
    def __init__(self, yaml_specification, task_repo, publication_repo, dst, langs,
                 html=False, join_langs=False, standalone=False,
                 css=None, header=None, translit=lambda x: x, babel=None):
        # call super constructors
        TaskVisitor.__init__(self)
        PublicationVisitor.__init__(self, yaml_specification, task_repo, publication_repo, langs=langs)

        # construct appropriate file writing mechanism (either a zip file, or an ordinary file system)
        if dst.endswith(".zip"):
            self._writer = ZipWriter(dst)
        else:
            self._writer = DirectoryWriter(dst)

        # construct appropriate Markdown serialization object (either
        # raw markdown or conversion to HTML)
        if html:
            self._md_serializer = HTMLMarkdownSerializer(self._writer, header=header, standalone=standalone, css=css, translit=translit, babel=babel)
        else:
            self._md_serializer = MarkdownSerializer(self._writer, header, translit=translit)

        # set the i18n
        if babel:
            messages.set_language(babel)
            
        # remember parameters
        self._join_langs = join_langs
        self._standalone = standalone

        # javascript language switcher id
        self._switcher_id = 0
        # javascript popup id
        self._popup_id = 0

        # tasks are not inline by default
        self._inline_task = False

        # dictionary that maps section paths to section titles (read from index.md files)
        self._section_titles = {}

        # construct the appropriate processors
        self._output_organizer = OutputOrganizerHierarchy()
        image_processor = ImageProcessorCopy(self._writer, self._output_organizer)
        link_processor = LinkProcessorHTML(self._yaml_specification, self._task_repo, self._output_organizer)
        self._md_processor = MDContentProcessor(link_processor, image_processor)

    # directory where html files for the current task will be stored
    # (taking into account the section in which that task occurs)
    def task_output_dir(self, task_id, section_path=None):
        if not section_path:
            section_path = self._extra_info["section_path"]

        # inline tasks go to the section path
        if self._inline_task:
            return self._output_organizer.output_dir(section_path)

        # regular tasks have their own dir
        return self._output_organizer.output_dir(os.path.join(section_path, task_id))
        
    # join solutions in all languages and add switcher if there is more than one language
    def join_languages(self, lang_md):
        self._switcher_id += 1
        # join languages
        joined_md = ""
        for lang in lang_md:
            joined_md += javascript.div(lang_md[lang], lang, self._switcher_id)
        # add switcher if there is more than one language
        if len(lang_md) > 1:
            joined_md = javascript.add_switcher(joined_md, self._langs, self._switcher_id)
        return joined_md
    
    def start(self):
        # open file serialization 
        self._md_serializer.open()

    def read_and_process_md(self, md_file_name, section_path, level):
        # path to the md file
        md_file_path = os.path.join(section_path, md_file_name)
        # read the content
        title, content = self._publication_repo.read_md_file(md_file_path)
        md = md_util.heading(title) + "\n\n" + content
        
        # process content and join langs if necessary
        if markdown_magic_comments.contains_block(md, "lang"):
            lang_md = self._md_processor.process(section_path, md_file_path, md,
                                                 level=level, langs=self._langs)
            joined_md = self.join_languages(lang_md)
            md = joined_md
        else:
            md = self._md_processor.process(section_path, md_file_path, md, level=level)

        # process included tasks
        def include_task(task_spec_lines):
            self._inline_task = True
            task_spec = yaml.safe_load("\n".join(task_spec_lines))
            task_result = self.task(section_path, level, task_spec["id"], task_spec)
            self._inline_task = False
            return task_result if task_result != None else ""
        md = markdown_magic_comments.replace_magic_comment_directives(md, "task", include_task)

        # process multichoice (abc) questions
        def process_abc(abc_spec_lines):
            abc_spec = yaml.safe_load("\n".join(abc_spec_lines))
            process_abc.question_id += 1
            return javascript.abc_question(abc_spec, process_abc.question_id)
        process_abc.question_id = 0
        md = markdown_magic_comments.replace_magic_comment_directives(md, "abc", process_abc)


        def process_javascript(javascript_spec_lines):
            def copy(file_name):
                logger.info("Copy:", file_name)
                dst_dir = self._output_organizer.output_dir(section_path)
                dst_path = os.path.join(dst_dir, file_name)
                src_path = os.path.join(os.path.dirname(md_file_path), file_name)
                self._writer.copy_file(src_path, dst_path)

            
            javascript_spec = yaml.safe_load("\n".join(javascript_spec_lines))
            if not "src" in javascript_spec:
                logger.error("No src in javascript specification")
                return ""
            
            copy(javascript_spec["src"])
            if "extra" in javascript_spec:
                if isinstance(javascript_spec["extra"], str):
                    copy(javascript_spec["extra"])
                else:
                    for file_name in javascript_spec["extra"]:
                        copy(file_name)

            height = int(javascript_spec["height"]) if "height" in javascript_spec else None
            title = javascript_spec["title"] if "title" in javascript_spec else ""
            iframe = javascript.iframe(javascript_spec["src"], title=title, height=height)
            if "popup" in javascript_spec:
                self._popup_id += 1
                return javascript.popup(msg("SHOW_APPLET"), iframe, self._popup_id)
            else:
                return iframe
            
        md = markdown_magic_comments.replace_magic_comment_directives(md, "javascript", process_javascript)

        return title, md

    def create_toc(self, section_path, md_files, subsections, tasks):
        md = ""

        def link_target(path):
            target_path = self._md_serializer.path(path)
            return os.path.relpath(target_path, section_path)

        # process md files
        for md_file in md_files:
            md_file_path = os.path.join(section_path, md_file)
            title, _ = self._publication_repo.read_md_file(md_file_path)
            md += md_util.list_item(md_util.link(title, link_target(md_file_path)))
        
        # process subsections
        if subsections:
            md += "\n{}:\n\n".format(md_util.bold(msg("SECTIONS")))
            for subsection in subsections:
                index_md = os.path.join(subsection, "index.md")
                title, _ = self._publication_repo.read_index(subsection)
                md += md_util.list_item(md_util.link(title, link_target(index_md)))
        if tasks:
            # process tasks
            md += "\n{}:\n\n".format(md_util.bold(msg("TASKS")))
            for task_id in tasks:
                task = self._task_repo.task(task_id)
                if not task:
                    logger.error(task_id, "not present in the task repository")
                    continue
                st_md_file = task.id() + "-st.md"
                task_dir = self.task_output_dir(task.id(), section_path)
                st_md_path = os.path.join(task_dir, st_md_file)
                md += md_util.list_item(md_util.link(task.title(), link_target(st_md_path)))
        return md

        
    # process the given publication source Markdown file
    def md_file(self, section_path, level, md_file_name, unnumbered=False):
        logger.info("Markdown file:", md_file_name)
        # read and process given markdown file
        title, md = self.read_and_process_md(md_file_name, section_path, level)
        # write the result to the file (converting to HTML if necessary)
        path = os.path.join(section_path, md_file_name)
        self._md_serializer.write(path, md, title=title)

    def section_start(self, section_path, level, md_files, subsections, tasks):
        logger.info("Section:", section_path)
        # process index.md for the section, if available        
        if not (self._publication_repo.contains_index(section_path)):
            logger.error(os.path.join(section_path, "index.md"), "does not exist in the publication repository")
            return
            
        # remember the title for this section
        title, _ = self._publication_repo.read_index(section_path)
        self._section_titles[section_path] = title

        # read and process the content
        title, md = self.read_and_process_md("index.md", section_path, level)
        # Make internal TOC for this section
        md += self.create_toc(section_path, md_files, subsections, tasks)
            
        # write the result to the file (converting to HTML if necessary)
        path = os.path.join(section_path, "index.md")
        self._md_serializer.write(path, md, title=title)


    def task_start(self, task):
        logger.info(task.id())
        # task statement in Markdown format
        self._st_md = ""
        
    def task_title(self, task):
        # append the title of the task as a heading
        self._st_md += md_util.heading(task.title()) + "\n\n"
        
    def task_st(self, task):
        current_occurrence = self._extra_info["current_occurrence"]
        # warn if this is a repeated task that it already occurred in the publication
        if current_occurrence > 1:
            self._st_md += md_util.italic(msg("REPEATED_TASK")) + "\n\n"
            
        # append the statement content, read from the task repository
        self._st_md += task.statement() + "\n\n"
        # process obtained md
        self._st_md = self._md_processor.process(self.task_output_dir(task.id()), task.st_path(), self._st_md)

    def task_io(self, task, description=True, examples=True):
        # read input-output specification
        if description and examples:
            md = task.io()
        elif description:
            md = task.io_description()
        elif examples:
            md = task.io_examples()
        # append it to the statement md
        self._st_md += task.io()
        # process obtained md
        self._st_md = self._md_processor.process(self.task_output_dir(task.id()), task.st_path(), self._st_md)        

    def task_end_st(self, task):
        # append the link to the solution
        if not self._inline_task and self._standalone and "solution" in self._what_to_print:
            self._st_md += "\n\n{}\n\n".format(md_util.link(md_util.italic(msg("SOLUTION")), self._md_serializer.path(task.id() + "-sol.md")))
            
    def solution_title(self, task, statement_link=True):
        # solution title
        title = md_util.heading("{} - {}".format(task.title(), msg("SOLUTION")))
        # add link to the problem statement
        if statement_link and self._standalone:
            title += "\n\n" + md_util.link(md_util.italic(msg("STATEMENT")), self._md_serializer.path(task.id() + "-st.md"))
        return title
            
    def task_sol(self, task, sols):
        # process the solution content separately for each language
        sol_md = task.sol_content()
        self._sol_md = self._md_processor.process(self.task_output_dir(task.id()), task.sol_path(), sol_md, langs=self._langs)

    def task_source_code(self, task, sol_name, sol_desc, lang, functions):
        # read the source code from the repository
        code = task.src_code(sol_name, lang)
        if not code:
            logger.error("missing code", task.id(), sol_name, lang)
            return
        # remove the magic comments (retain the whole source code)
        code = source_code_magic_comments.remove_magic_comments(code)
        # extract just specific functions
        if functions != None:
            code = code_parser.extract_funs(lang, code, functions)

        # surround it with Markdown markup for program source code
        code = "\n" + md_util.md_source_code(code, lang)
        # insert the code to appropriate place in the solution for its language
        if "solution" in self._what_to_print:
            self._sol_md[lang] = markdown_magic_comments.insert_content(self._sol_md[lang], "sol", sol_name, code, "code", "here")
        else:
            if not hasattr(self, "_sol_md"):
                self._sol_md = dict()
            self._sol_md[lang] = code

    def task_report_pending_occurrences(self, task, current_occurrence, total_occurrences):
        # notify that the problem has other solutions
        md = md_util.italic(msg("ADDITIONAL_SOLUTIONS_EXIST"))
        self._st_md += md + "\n\n"

    def task_result(self, task):
        # join solutions in all languages and add switcher
        joined_sol_md = ""
        if "statement" in self._what_to_print:
            joined_sol_md += md_util.heading(msg("SOLUTION"), 2) + "\n"
        joined_sol_md += self.join_languages(self._sol_md)
        return self._st_md + "\n" + joined_sol_md
        
    def task_end(self, task):
        current_occurrence = self._extra_info["current_occurrence"]
        total_occurrences = self._extra_info["total_occurrences"]
        if current_occurrence < total_occurrences:
            self.task_report_pending_occurrences(task, current_occurrence, total_occurrences)

        if not self._inline_task:
            # write statement to file
            st_path = os.path.join(self.task_output_dir(task.id()), task.id() + "-st.md")
            self._md_serializer.write(st_path, self._st_md, title=task.title(), md_time=task.modification_time())

            if "solution" in self._what_to_print:
                title = "{} - {}".format(task.title(), msg("SOLUTION"))
                if not self._join_langs:
                    # write solutions in every language into separate files (converting to HTML if necessary)
                    for lang in self._sol_md:
                        # add title
                        sol_md = self.solution_title(task) + "\n\n"
                        # add solution in the current language
                        sol_md += self._sol_md[lang]
                        # write solution to file (converting to HTML if necessary)
                        sol_path = os.path.join(self.task_output_dir(task.id()), "{}-{}-sol.md".format(task.id(), lang))
                        self._md_serializer.write(sol_path, sol_md, title=title, md_time=task.modification_time())
                else:
                    # add title
                    joined_sol_md = self.solution_title(task) + "\n\n"
                    # join solutions in all languages and add switcher
                    joined_sol_md += self.join_languages(self._sol_md)
                    # write the result into a single solution file (converting to HTML if necessary)
                    sol_path = os.path.join(self.task_output_dir(task.id()), "{}-sol.md".format(task.id()))
                    self._md_serializer.write(sol_path, joined_sol_md, title=title, md_time=task.modification_time())
            
    # action called at the end of publication processing 
    def end(self):
        pass
